#!/usr/bin/env/python
# Tells Linux systems how to run this file.

"""
Toomi
Socket-based server tool
"""

__author__ = "Kris Schnee"
__version__ = "2009.11.7"
__license__ = "GPLv3 or MIT"

##### Imported Modules #####

## Standard

import socket
import select
import time

## Third-party


## Mine


##### Constants #####

HOST = "localhost"
PORT = 7777

##### Classes #####
        
class Server:
    def __init__(self,internet_host=True,port=PORT):
        self.clients = {}

        self.listen_socket = socket.socket( socket.AF_INET, socket.SOCK_STREAM )
        self.listen_socket.setblocking(False)
        if internet_host:
            host = socket.gethostbyname(socket.gethostname())
        else:
            host = "localhost"
        self.listen_socket.bind( (host, port) )
        self.listen_socket.listen(4)
        print "Server started. My IP address and port: "+str(host)+":"+str(port)

    def ShutDownServer(self,end_program=True):
        self.listen_socket.close()
        if end_program:
            exit()

    def HandleInput(self,address,data):
        """Override this."""
        result = "Input received: "+data
        return result

    def CheckForConnections(self):
        """Check for connection attempts."""
        rejoiners = [] ## Known addresses, reconnecting for some reason.
        newcomers = []
        
        a = [self.listen_socket]
        b = []
        c = []
        readable, writeable, erroneous = select.select(a,b,c,.001)
        if readable:
            print "Listen socket seems to be readable. Checking now."
            socket, address = self.listen_socket.accept()
            print "Got a connection from "+str(address)
            socket.setblocking(False)
            
            if address in self.clients:
                print "Rejoining client: "+str(address)
                rejoiners.append((socket,address))
            else:
                print "New client: "+str(address)
                newcomers.append((socket,address))
        return rejoiners, newcomers

    def HandleNewConnections(self,rejoiners=[],newcomers=[]):
        """React to newly connected clients. Override this."""

        for socket, address in rejoiners:
            self.clients[address]["socket"].close() ## Close old socket
            self.clients[address]["socket"] = socket ## Replace socket
            self.clients[address]["last_input"] = int(time.time)

        for socket, address in newcomers:
            self.clients[address] = {"last_input":int(time.time()),
                                     "name":"stranger",
                                     "socket":socket,
                                     }

    def CheckForInput(self):
        """Get input from clients already connected."""
        inputs = {}
        a = [c["socket"] for c in self.clients.values()]
        b = []
        c = []
        if a:
            readable, writeable, erroneous = select.select(a,b,c,.001)
            for client in self.clients:
                s = self.clients[client]["socket"]
                if s in readable:
                    ## What you say?
                    data = s.recv(4096).replace("\r","")
                    ## Certain endline characters, "\r", are stripped.
                    inputs[client] = data
        return inputs

    def DemoLoop(self):
        while True:
            rejoiners, newcomers = self.CheckForConnections()
            if rejoiners or newcomers:
                self.HandleNewConnections(rejoiners,newcomers)
                
            inputs = self.CheckForInput()
            for address in inputs:
                if inputs[address]:
                    ## Act based on the input.
                    result = self.HandleInput(address,inputs[address])
                    self.clients[address]["socket"].send(result)
            time.sleep(.1)




class Client:
    def __init__(self):
        self.socket = None

    def Connect(self,host=HOST,port=PORT):
        self.socket = socket.socket( socket.AF_INET, socket.SOCK_STREAM )
        self.socket.connect( (HOST, PORT) )
        self.socket.setblocking(0)

    def Disconnect(self):
        self.socket.close()

    def Send(self,output=""):
        """Open a socket connection and send data to the server."""
        if output:
            self.socket.send(output)

    def ListenToServer(self):
        """Get input."""
        readable, writeable, erroneous = select.select([self.socket],[],[])
        if readable:
            data = self.socket.recv(4096)
            return data

    def DemoLoop(self):
        print "Starting client program."
        self.Connect()
        while True:
            output = raw_input("Send what? > ")

            if output.lower() in ["q","quit","exit"]:
                print "Quitting."
                break
            
            self.SendToServer(output)
            reply = self.ListenToServer()
            print reply



class ServerDemo( Server ):
    def __init__(self,use_internet=True,port=PORT):
        Server.__init__(self,use_internet,port)
        self.weather = "sunny"

    def HandleInput(self,address,data):
        result = "Input received: "+data

        if data.lower() == "shutdown":
            self.ShutDownServer()

        words = data.split(" ")
        if words:
            if words[0] == "weather":
                self.weather = words[1]

            elif words[0] == "name":
                self.clients[address]["name"] = " ".join(words[1:])

        name = self.clients[address]["name"]

        result = "Hello, "+name+". The weather is currently "+self.weather+".\nMy currently connected clients are: "
        result += ", ".join([c["name"] for c in self.clients.values()])

        return result



##### Other Functions #####

def AskForIP():
    host = raw_input("Host IP address? (leave blank for localhost) > ")
    if not host:
        host = HOST
    port = raw_input("Host port? (leave blank for default) > ")
    if not port:
        port = PORT
    return host, port

def AskUseInternet():
    x = raw_input("Start an Internet server? Type 'y' for yes, 'n' to run a 'localhost' server on this machine only. > ")
    if x in "yY":
        return True
    return False

def GetMyIP():
    return socket.gethostbyname(socket.gethostname())

##### Autorun #####

if __name__ == "__main__":
    ## Runs if this module is the main module (not imported).
    use_internet = AskUseInternet()
    s = ServerDemo(use_internet)
    s.DemoLoop()
